enum CbcTagId {
  CBC_TAG_BYTE_ORDER,
  CBC_TAG_DIMENSION,
  CBC_TAG_FORMAT,
  CBC_TAG_HOOKS,
  CBC_INVALID_TAG
};

enum CbcTagByteOrder {
  CBC_TAG_BYTE_ORDER_BIG_ENDIAN,
  CBC_TAG_BYTE_ORDER_LITTLE_ENDIAN,
  CBC_INVALID_BYTE_ORDER
};

enum CbcTagFormat {
  CBC_TAG_FORMAT_STRING,
  CBC_TAG_FORMAT_BINARY,
  CBC_INVALID_FORMAT
};
